import os, time
from StringIO import StringIO
import subprocess, shlex
import whatap.util.logging_util as logging_util
import re


filesystems = { "btrfs": True, "ext2": True, "ext3": True, "ext4": True, "reiser": True, "xfs": True,
        "ffs": True, "ufs": True, "jfs": True, "jfs2": True, "vxfs": True, "hfs": True, "ntfs": True, "fat32": True,
        "zfs": True, "refs": True, "nfs": True, "nfs2": True, "nfs3": True, "nfs4": True, "cifs": True, "ocfs2": True }

zfsInfoCheck = False

class _PoolInfo:
    def __init__(self, name):
        self.name = name
        self.totalSpace = None
        self.diskList = []
        self.ioInfo = _IoInfo()
        self.oldTimeStamp = 0
        self.readIops = 0
        self.writeIops = 0
        self.readBps = 0
        self.writeBps = 0
        self.ioPercent = 0


    def setTotalSpace(self, totalSpace):
        self.totalSpace = totalSpace

    def setDisk(self, disk):
        self.diskList.append(disk)

    def initIoInfo(self):
        self.ioInfo.writeIoByte = 0
        self.ioInfo.writeIoCount = 0 
        self.ioInfo.readIoByte = 0
        self.ioInfo.readIoCount = 0 
        self.ioInfo.ioMillis = 0 

    def sumIoStats(self, newInfo):
        self.ioInfo.writeIoByte += newInfo.writeIoByte
        self.ioInfo.writeIoCount += newInfo.writeIoCount
        self.ioInfo.readIoByte += newInfo.readIoByte
        self.ioInfo.readIoCount += newInfo.readIoCount
        if self.ioInfo.ioMillis < newInfo.ioMillis:
            self.ioInfo.ioMillis = newInfo.ioMillis

    def calcIoStat(self, diffTime):
        self.writeBps = float(self.ioInfo.writeIoByte) / float(diffTime)
        self.writeIops = float(self.ioInfo.writeIoCount) / float(diffTime)
        self.readBps = float(self.ioInfo.readIoByte) / float(diffTime)
        self.readIops = float(self.ioInfo.readIoCount) / float(diffTime)
        self.ioPercent = float(self.ioInfo.ioMillis) / float(diffTime * 1000) * 100
        if self.ioPercent > float(100):
            self.ioPercent = float(100)
        self.initIoInfo()

            
 
class _DataSetInfo:
    def __init__(self, name):
        self.name = name
        self.pool = None
        self.used = 0
        self.avail = 0
        self.recordSize = 0

    def setDataSet(self, pool, used, avail, recordSize):
        self.pool = pool
        self.used = used
        self.avail = avail
        self.recordSize = recordSize

    def setRecodeSize(self, recordSize):
        self.recordSize = recordSize


class _IoInfo:
    def __init__(self):
        self.writeIoByte= 0
        self.writeIoCount = 0
        self.readIoByte= 0
        self.readIoCount = 0
        self.ioMillis= 0
        self.timestamp = 0

    def setIoInfo(self, writeIoByte, writeIoCount, readIoByte, readIoCount, ioMillis, timestamp):
        self.writeIoByte= writeIoByte
        self.writeIoCount = writeIoCount
        self.readIoByte= readIoByte
        self.readIoCount = readIoCount
        self.ioMillis = ioMillis
        self.timestamp = timestamp


def getDiskList():
    diskList = list(set(os.listdir("/dev/rdsk") + os.listdir("/dev/dsk")))
    return diskList

oldZfsDiskList = None
diskToPoolMap = None

def getZpoolInfo():
    global oldZfsDiskList
    diskList = getDiskList()

    if diskList == oldZfsDiskList:
        return None

    command = "zpool list -H -o name,size"
    try:
        process = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        text = process.stdout.read()

        info_map = {}
        for line in text.strip('\n').split('\n'):
            name, size = line.split('\t')
            pool = _PoolInfo(name)
            pool.setTotalSpace(size)
            info_map[name] = pool

        info_map = getZpoolStatus(info_map, diskList)
        oldZfsDiskList = diskList
        return info_map
    except subprocess.CalledProcessError as e:
        logging_util.debug("zpool list error : ", e)
        return None


def getZpoolStatus(info_map, diskList):
    diskMap = {}
    poolMap = {}
    global diskToPoolMap

    #command = "paste -d, <(iostat -x | awk 'NR>2{print $1}') <(iostat -nx | awk 'NR>2{print $11}')"
    command = "tmp1=/tmp/tmpfile1 tmp2=/tmp/tmpfile2 && iostat -x | awk 'NR>2{print $1}' > $tmp1 && iostat -nx | awk 'NR>2{print $11}' > $tmp2 && paste -d, $tmp1 $tmp2 && rm $tmp1 $tmp2"
    try:
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        text = process.stdout.read()
        for line in text.strip('\n').split('\n'):
            deviceName, deviceRealName = line.split(',')
            diskMap[deviceRealName] = deviceName

    except subprocess.CalledProcessError as e:
        logging_util.debug("io error : ", e)
        return None

    command = "zpool status"
    try:
        process = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        text = process.stdout.read().strip('\n')
        poolKey = ""
        
        for line in re.split(r'[\n\t]+', text):
            status = [value for value in line.split(' ') if value]
            if status[0] in info_map.keys():
                poolKey = status[0]
            elif status[0] in diskList and poolKey in info_map:
                if poolKey == "": continue
                else: 
                    diskName = re.sub(r's\d+', '', status[0])
                    if diskName in diskMap:
                        info_map[poolKey].setDisk(diskName)
                        poolMap[diskMap[diskName]] = poolKey
        diskToPoolMap = poolMap
        return info_map

    except subprocess.CalledProcessError as e:
        logging_util.debug("zpool status error : ", e)
        return None



def diffIoInfo(total, old):
    if total == None or old == None:
        return None
    ret = _IoInfo()
    ret.setIoInfo(total.writeIoByte - old.writeIoByte, total.writeIoCount- old.writeIoCount, total.readIoByte - old.readIoByte, total.readIoCount - old.readIoCount, total.ioMillis - old.ioMillis, total.timestamp - old.timestamp)
    return ret

oldIoInfoMap = None
import psutil
def getDiskIoInfo():
    ioStats = psutil.disk_io_counters()
    info = _IoInfo()
    now = time.time() 

    info.setIoInfo(ioStats.write_bytes, ioStats.write_count, ioStats.read_bytes, ioStats.read_count, ioStats.write_time + ioStats.read_time, now)

    global oldIoInfoMap
    if not oldIoInfoMap:
        oldIoInfoMap = info
        return

    ret = diffIoInfo(info, oldIoInfoMap)
    return ret
    

def getDiskIoInfoPerDisk(diff):
    ioStats = psutil.disk_io_counters(perdisk=True)
    ioInfoMap = {}
    now = time.time() 

    for k, v in ioStats.items():
        info  = _IoInfo()
        info.setIoInfo(v.write_bytes, v.write_count, v.read_bytes, v.read_count, v.write_time+ v.read_time, now)
        ioInfoMap[k] = info

    global oldIoInfoMap 
    if not oldIoInfoMap:
        oldIoInfoMap = ioInfoMap
        return

    for k, v in ioInfoMap.items():
        diffIo = diffIoInfo(v, oldIoInfoMap[k])
        if not diffIo:
            continue
        try:
            zpoolInfoMap[diskToPoolMap[k]].sumIoStats(diffIo)
        except KeyError, e:
            logging_util.debug("key error : ", e)
            continue

    oldIoInfoMap = ioInfoMap
    for k, v in zpoolInfoMap.items():
        v.calcIoStat(diff)



def getZfsInfo():
    try:
        command = "zfs list -H -o name,used,avail,recordsize"
        process = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        text = process.stdout.read()
        
        info_map = {}
        for line in text.strip('\n').split('\n'):
            name, used, avail, recordSize = line.split('\t')
            pool = name.split('/')[0]
            dataSet = _DataSetInfo(name)
            dataSet.setDataSet(pool, used, avail, recordSize)
            info_map[name] = dataSet
        return info_map
    except subprocess.CalledProcessError as e:
        logging_util.debug("zfs list error : ", e)
        return None


zpoolInfoMap = getZpoolInfo()
zfsInfoMap = getZfsInfo()
oldTime = time.time()

def setZfsInfo():
    global oldTime
    now = time.time()

    global zpoolInfoMap 
    poolCheck = getZpoolInfo()
    if poolCheck != None:
        zpoolInfoMap = poolCheck

    global zfsInfoMap 
    zfsInfoMap = getZfsInfo()
    diffTime = now - oldTime
    getDiskIoInfoPerDisk(diffTime)

    oldTime = now


class _DiskPerf:
    def __init__(self):
        self.mountPoint = None
        self.deviceId = None
        self.fileSystem = None
        self.totalSpace = None
        self.usedSpace = None
        self.usedPercent = None
        self.freeSpace = None
        self.freePercent = None
        self.blksize = None
        self.readIops = 0
        self.writeIops = 0
        self.readBps = 0
        self.writeBps = 0
        self.ioPercent = 0
        self.queueLength = 0

    def __str__(self):
        buf = StringIO()
        buf.write('mountPoint=')
        buf.write(self.mountPoint) 
        buf.write('\ndeviceId=')
        buf.write(self.deviceId)
        buf.write('\nfileSystem=')
        buf.write(self.fileSystem)
        buf.write('\ntotalSpace=')
        buf.write(self.totalSpace)
        buf.write('\nusedSpace=')
        buf.write(self.usedSpace)
        buf.write('\nusedPercent=')
        buf.write(self.usedPercent)
        buf.write('\nfreeSpace=')
        buf.write(self.freeSpace)
        buf.write('\nfreePercent=')
        buf.write(self.freePercent)
        buf.write('\nself.blksize=')
        buf.write(self.blksize)
        buf.write('\nreadIops=')
        buf.write(self.readIops)
        buf.write('\nwriteIops=')
        buf.write(self.writeIops)
        buf.write('\nreadBps=')
        buf.write(self.readBps)
        buf.write('\nwriteBps=')
        buf.write(self.writeBps)
        buf.write('\nioPercent=')
        buf.write(self.ioPercent)
        buf.write('\nqueueLength=')
        buf.write(self.queueLength)

        return buf.getvalue()

def parse_byte_value(byte_str):
    if byte_str.endswith('K'):
        return int(float(byte_str[:-1]) * 1024)
    elif byte_str.endswith('M'):
        return int(float(byte_str[:-1]) * 1024 * 1024)
    elif byte_str.endswith('G'):
        return int(float(byte_str[:-1]) * 1024 * 1024 * 1024)
    elif byte_str.endswith('T'):
        return int(float(byte_str[:-1]) * 1024 * 1024 * 1024 * 1024)
    else:
        return int(byte_str)

def parse_count_value(count_str):
    if count_str.endswith('K'):
        return int(float(count_str[:-1]) * 1000)
    elif count_str.endswith('M'):
        return int(float(count_str[:-1]) * 1000 * 1000)
    elif count_str.endswith('G'):
        return int(float(count_str[:-1]) * 1000 * 1000 * 1000)
    elif count_str.endswith('T'):
        return int(float(count_str[:-1]) * 1000 * 1000 * 1000 * 1000)
    else:
        return int(count_str)



def getZFSDisk(diskPerf):
    if zfsInfoMap is None or zpoolInfoMap is None:
        return diskPerf

    dataSetInfo = zfsInfoMap[diskPerf.deviceId]
    poolInfo = zpoolInfoMap[dataSetInfo.pool]

    diskPerf.totalSpace = parse_byte_value(poolInfo.totalSpace)
    diskPerf.usedSpace = parse_byte_value(dataSetInfo.used)
    diskPerf.usedPercent = float(100.0 * float(diskPerf.usedSpace) / float(diskPerf.totalSpace))
    diskPerf.freeSpace = parse_byte_value(dataSetInfo.avail)
    diskPerf.freePercent = float(100.0 * float(diskPerf.freeSpace) / float(diskPerf.totalSpace))
    diskPerf.blksize = int(parse_byte_value(dataSetInfo.recordSize))
    diskPerf.readBps = poolInfo.readBps 
    diskPerf.writeBps = poolInfo.writeBps
    diskPerf.readIops = poolInfo.readIops
    diskPerf.writeIops = poolInfo.writeIops
    diskPerf.ioPercent = poolInfo.ioPercent
    return diskPerf


def getDisk():
    f = open('/etc/mnttab','r')
    if not f:
        return None
    try:
        global zfsInfoCheck
        diskQuotas = []
        for l in f.readlines():
            words = l.split()
            if words[2] in filesystems:
                deviceId = words[0]
                p = _DiskPerf()
                p.mountPoint = words[1]
                p.deviceId = deviceId
                p.fileSystem = words[2]

                if words[2] == 'zfs' or words[2] == 'vzfs':
                    if zfsInfoCheck == False:
                        zfsInfoCheck = True
                        setZfsInfo()
                    p = getZFSDisk(p)
                    diskQuotas.append(p)
            
                else:
                    f_bsize, f_frsize, f_blocks, f_bfree, f_bavail, f_files, f_ffree, f_favail, f_flag, f_namemax = os.statvfs(p.mountPoint)
                    if f_blocks == 0:
                        continue
                    p.totalSpace = int(f_frsize * f_blocks)
                    p.usedSpace = int(f_frsize * (f_blocks - f_bfree))
                    p.freeSpace = p.totalSpace - p.usedSpace
                    p.usedPercent = float(100.0 * float(f_blocks - f_bfree) / float(f_blocks))
                    p.freePercent = float(100.0 * f_bfree / f_blocks)
                    p.blksize = int(f_frsize)
                    diskQuotas.append(p)
    finally:
        f.close()

    zfsInfoCheck = False
    return diskQuotas

def test():
    for a in getDisk():
        print a
    time.sleep(10)
    for a in getDisk():
        print a

if __name__ == '__main__':
    test()

